"
COPYRIGHT (c) 2020 LabWare
"
"{ Package: 'stx:goodies/monticello/tonel' }"

"{ NameSpace: Smalltalk }"

Object subclass:#TonelParser
	instanceVariableNames:'stream'
	classVariableNames:''
	poolDictionaries:''
	category:'SCM-Monticello-Tonel-Core'
!

TonelParser comment:'I''m a parser for tonel files. 
I parse a class with the following format: 

Tonel spec
====

    [comment]
    type { typeDefinition }
    (
        [{ methodMetadata }]
        method [
            methodBody ] 
    )*


comment
---
"
comment string
"
is optional (but it should be there, in good design ;)

type
---
Class|Trait|Extension

typeDefinition
---
a STON file with class/trait/extension metadata

methodMetadata
---
a STON file with method metadata
is optional (but also, recommended)

method
---
method declaration as this: 

Class[ class] >> selector

methodBody 
---
the method body (we do not parse contents, that''s class builder task)'
!

!TonelParser class methodsFor:'documentation'!

copyright
"
COPYRIGHT (c) 2020 LabWare

"
! !

!TonelParser class methodsFor:'instance creation'!

on: aStream 
    ^ self new 
        stream: aStream;
        yourself
! !

!TonelParser class methodsFor:'parsing'!

parseStream: aStream 
    ^ (self on: aStream) start
!

parseString: aString 
    ^ self parseStream: aString readStream
! !

!TonelParser methodsFor:'accessing'!

start
    ^ self document
!

stream: aStream 
    stream := aStream
! !

!TonelParser methodsFor:'parsing'!

comment
    ^ (TonelCommentScanner on: stream) scan
!

document
    "
    ^ { 
    self typeDef.
    self methodDefList.
     } 
    flattened
    select: #notNil
    "
    ^ ((Array with: self typeDef) , self methodDefList) select: [:each | each notNil ]

    "Modified: / 07-07-2020 / 22:44:10 / Jan Vrany <jan.vrany@labware.com>"
!

metadata
    | result ch count |
    
    result := String new writeStream.

    count := 0.
    stream peek = ${ ifFalse: [ TonelParseError signal: 'Can''t parse metadata' ].	
    [ stream atEnd ]
    whileFalse: [ 
        ch := stream next.
        result nextPut: ch.
        ch = ${ ifTrue: [ count := count +1 ].
        ch = $} ifTrue: [ count := count -1 ].
        count = 0 ifTrue: [ ^ STON fromString: result contents ]].

    TonelParseError signal: 'Can''t parse metadata'
!

method
    | type selector |
    
    type := self untilIncluding: '>>'.
    selector := self cleanSelector: (self untilExcluding: '[').
    type := (type withoutSeparators substrings: ' ') asArray.
    type size = 1 ifTrue: [ type := type copyWith: nil ].
    ^ { 
        type.
        selector.
    }

    "Modified: / 25-05-2020 / 14:34:32 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 27-05-2020 / 16:02:40 / Jan Vrany <jan.vrany@labware.com>"
!

methodBody
    "I read a methodbody (what is inside [ ... ])"
    ^ (TonelSourceScanner on: stream) scan
        
    
!

methodDef
    ^ self newMethodDefinitionFrom: { 
        self separator.
        self try: [ self metadata ]. 
        self separator. 
        self method. 
        self methodBody 
    }
!

methodDefList
    | result |
    
    self separator. "to arrive to the end of the file in case there are no methods"
    result := Array new writeStream.
    [ stream atEnd ]
    whileFalse: [ 
        result nextPut: self methodDef.
        "skip possible spaces at the end"
        self separator ].
    
    ^ result contents
!

separator
    [ stream atEnd not and: [ self isSeparator: stream peek ] ]
    whileTrue: [ stream next ].
    ^ nil
!

type
    self try: [ self word: 'Class' ] onSuccess: [ :word | ^ word  ].
    self try: [ self word: 'Trait' ] onSuccess: [ :word | ^ word  ].
    self try: [ self word: 'Extension' ] onSuccess: [ :word | ^ word  ].
    
    "at end"
    TonelParseError signal: 'Can''t parse type.'	
!

typeDef
    ^ self newTypeDefinitionFrom: { 
        self separator.
        self try: [ self comment ]. 
        self separator. 
        self type. 
        self separator. 
        self try: [ self metadata ] 
    } 
    
        
! !

!TonelParser methodsFor:'private'!

cleanSelector: aString
    "BEWARE: I'm doing some heave assumptions here: I'm removing just ONE space (in case there 
     is one) because I expect this to be a file generated by tonel, and tonel adds one space 
     before start with the method body to make the format more readable. 
     But of course this is not very good :("
    ^ (aString last = Character space
        ifTrue: [ aString allButLast ]
        ifFalse: [ aString ]) 
        trimLeft
!

extractSelector: aString
    | selectorStream keywords |

    keywords := Array new writeStream.
    selectorStream := (self removeComments: aString) readStream.

    [ selectorStream atEnd ] whileFalse: [ 
        | word ch |

        selectorStream skipSeparators.
        word := String new writeStream.
        [ selectorStream atEnd not and: [ (ch := selectorStream next) isSeparator not and:[ch ~~ $: ] ] ] whileTrue: [
            word nextPut: ch 
        ].
        ch = $: ifTrue: [ word nextPut: ch ].
        keywords nextPut: word contents 
    ] .
    keywords := keywords contents.

    ^ (keywords size <= 2 
        ifTrue: [ keywords first]
        ifFalse: [ (keywords pairWiseCollect: [ :keyword :argument | keyword ]) asStringWith:'' ])
        asSymbol

    "
    TonelParser new extractSelector: 'some:         arg1 keyword:arg2'
    "

    "Modified: / 25-05-2020 / 14:41:07 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 27-05-2020 / 15:55:08 / Jan Vrany <jan.vrany@labware.com>"
!

removeComments: original
    | newStream readStream inComment |
    
    newStream := original copy writeStream.
    readStream := original readStream.
    inComment := false.
    
    [ readStream atEnd ] whileFalse: [ | ch |
        
        ch := readStream next.
        
        (ch = $") ifTrue:[
            inComment := inComment not.
            ch := readStream next.
        ].
        
        (inComment or:[ ch isNil]) ifFalse: [ newStream nextPut: ch	 ]
    ].
    
    ^ newStream contents
!

validateStatelessTraitIsBeingRead: metadata
    | vars |
    
    vars := Set new.
    vars 
        addAll: (metadata at: #instVars ifAbsent: [ #() ]);
        addAll: (metadata at: #classInstVars ifAbsent: [ #() ]).
    
    vars ifNotEmpty: [ 
        self error: 'Trying to load a stateful trait in a stateless version.' ]
! !

!TonelParser methodsFor:'private factory'!

definitionForType: aString 
    aString = TonelWriter classLabel ifTrue: [ ^ MCClassDefinition ].
    aString = TonelWriter traitLabel ifTrue: [ ^ MCTraitDefinition ].	
    aString = TonelWriter extensionLabel ifTrue: [ ^ nil ].
    
    TonelParseError signal: 'Unknown type declaration.'
!

newClassDefinitionFrom: anArray
    | metadata |

    metadata := anArray sixth.
    ^ MCClassDefinition
        name: (metadata at: #name)
        superclassName: (metadata at: #superclass)
        traitComposition: (metadata at: #traits ifAbsent: [ '{}' ])
        classTraitComposition: (metadata at: #classTraits ifAbsent: [ '{}' ])
        category: (metadata at: #category)
        instVarNames: (metadata at: #instVars ifAbsent: [ #() ])
        classVarNames: (metadata at: 'classVars' ifAbsent: [ #() ])
        poolDictionaryNames: (metadata at: 'pools' ifAbsent: [ #() ])
        classInstVarNames: (metadata at: 'classInstVars' ifAbsent: [ #() ])
        type: (metadata at: #type ifAbsent: [ #normal ]) asSymbol
        comment: (anArray second ifNil: [ '' ])
        commentStamp: nil
!

newMethodDefinitionFrom: anArray
    | metadata className meta selector source  |
    
    metadata := anArray second ifNil: [ Dictionary new ].
    className := anArray fourth first first.
    meta := anArray fourth first second notNil.
    selector := self extractSelector: anArray fourth second withoutSeparators.
    source := String streamContents: [ :s | 
        s nextPutAll: anArray fourth second.
        anArray fifth ifNotEmpty: [ :src | s nextPutAll: src ] ].
    
    ^ (MCMethodDefinition
        className: className
        classIsMeta: meta
        selector: selector
        category: (metadata at: #category ifAbsent: [ '' ]) 
        timeStamp: nil
        source: source)
        source: source "/ <-- this is to make sure line endings are preserved

    "Modified: / 25-05-2020 / 14:34:36 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 07-07-2020 / 23:34:05 / Jan Vrany <jan.vrany@labware.com>"
!

newStatefulTraitDefinitionFrom: anArray
    | metadata |
    
    metadata := anArray sixth.
    ^ MCTraitDefinition
            name: (metadata at: #name)
            traitComposition: (metadata at: #traits ifAbsent: [ '{}' ])
            category: (metadata at: #category)
            instVarNames: (metadata at: #instVars ifAbsent: [ #() ])
            classInstVarNames: (metadata at: #classInstVars ifAbsent: [ #() ])
            comment: (anArray second ifNil: [ '' ])
            commentStamp: nil
!

newStatelessTraitDefinitionFrom: anArray
    | metadata |
    
    metadata := anArray sixth.
    
    self validateStatelessTraitIsBeingRead: metadata.
    
    ^ MCTraitDefinition
            name: (metadata at: #name)
            traitComposition: (metadata at: #traits ifAbsent: [ '{}' ])
            category: (metadata at: #category)
            comment: (anArray second ifNil: [ '' ])
            commentStamp: nil
!

newTraitDefinitionFrom: anArray
    | metadata traitDefs |
    
    metadata := anArray sixth.
    traitDefs := { 
        self hasStatefulTraits 
            ifTrue: [ self newStatefulTraitDefinitionFrom: anArray ]
            ifFalse: [ self newStatelessTraitDefinitionFrom: anArray ] }.
            
    metadata 
        at: #classTraits
        ifPresent: [ :classTraits |
            traitDefs := traitDefs copyWith: (MCClassTraitDefinition
 				baseTraitName: (metadata at: #name)
 				classTraitComposition: classTraits
                category: (metadata at: #category)) ].
            
    ^ traitDefs
!

newTypeDefinitionFrom: anArray
    | typeClass |
    
    typeClass := self definitionForType: anArray fourth.
    typeClass = MCClassDefinition ifTrue: [ ^ self newClassDefinitionFrom: anArray ].
    typeClass = MCTraitDefinition ifTrue: [ ^ self newTraitDefinitionFrom: anArray ].	
    "is extension, no type"
    ^ nil
! !

!TonelParser methodsFor:'private parsing'!

try: aBlock
    ^ self 
        try: aBlock 
        onSuccess: [ :parsedValue | parsedValue ] 
        onFailure: [ nil ]
!

try: aBlock onFailure: failureBlock
    ^ self 
        try: aBlock 
        onSuccess: [ :parsedValue |  parsedValue ] 
        onFailure: failureBlock
!

try: aBlock onSuccess: successBlock
    ^ self 
        try: aBlock 
        onSuccess: successBlock 
        onFailure: [ nil ]
!

try: aBlock onSuccess: successBlock onFailure: failureBlock
    | pos |
    
    pos := stream position.
    [ ^ successBlock value: aBlock value ]
    on: TonelParseError 
    do: [ :e | 
        stream position: pos.
        ^ failureBlock value ]. 
    
!

untilExcluding: aCollection
    | result |
    result := stream upToAll: aCollection.
    stream position: stream position - aCollection size + 1.
    ^ result

    "Modified: / 28-05-2020 / 12:13:40 / Jan Vrany <jan.vrany@labware.com>"
!

untilIncluding: aCollection
    ^ stream upToAllExcluding: aCollection

    "Modified: / 27-05-2020 / 16:00:34 / Jan Vrany <jan.vrany@labware.com>"
!

word: aString
    | result |
    result := stream next: aString size.
    result = aString
        ifFalse: [ TonelParseError signal: 'Can''t parse ', aString ].
    ^ result
! !

!TonelParser methodsFor:'private testing'!

hasStatefulTraits
    "Pharo has stateful traits starting on version 7"
    
    ^ SystemVersion current major >= 7
!

isSeparator: aCharacter 
    ^ aCharacter isSeparator
! !

!TonelParser class methodsFor:'documentation'!

version_HG

    ^ '$Changeset: <not expanded> $'
! !

